/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 * 
 * TsmSupervisor.cs
 * version 32
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Text;
using System.Threading;
using System.Runtime.Remoting.Contexts;
using System.Diagnostics; // Process

namespace TsmRTE
{
  public enum TsmScenItemKind
  {
    RELEASE_CB_MSG,
    RELEASE_CB_RV2,
    RELEASE_MSG,
    RELEASE_RV2,
    NATIVE,
    CHOICE,
    QO
  }

  public class TsmScenItem
  {
    public TsmScenItemKind  Kind;
    public int              Index;
    public TsmClass         Receiver;
    public TsmThreadData    ThreadData;

    public override string ToString()
    {
      switch ( Kind )
      {
      case TsmScenItemKind.RELEASE_CB_MSG :
        return "CB_MSG";
      case TsmScenItemKind.RELEASE_CB_RV2 :
        return "CB_RV2";
      case TsmScenItemKind.RELEASE_MSG :
        return "MSG " + ThreadData.ThreadId;
      case TsmScenItemKind.RELEASE_RV2 :
        return "RV2 " + ThreadData.ThreadId;
      case TsmScenItemKind.NATIVE :
        return string.Format( "NATIVE{0,5} ", Index ) + Receiver.FullName();
      case TsmScenItemKind.CHOICE :
        return "CHOICE " + Index;
      case TsmScenItemKind.QO :
        return "QO";
      }
      return "TismBUG";
    }
  }

  public class TsmScenario
  {
    public static int         ScenarioId = 0;
    public int                Id;
    public TsmScenario        Parent;
    public List<TsmScenItem>  Items = new List<TsmScenItem>();
    public TsmSnapshot        Snapshot = null;

    public TsmScenario()
    {
      ScenarioId = 0;
      Id = ScenarioId++;
      Parent = null;
    }

    public TsmScenario( TsmScenario oParent )
    {
      Id = ScenarioId++;
      Parent = oParent;
    }

    public int ParentId
    {
      get { return ( Parent == null ) ? -1 : Parent.Id; }
    }

    public TsmScenario Extend( TsmScenItem oTsmScenItem )
    {
      TsmScenario lTsmScenario = new TsmScenario( this );

      lTsmScenario.Items = new List<TsmScenItem>( this.Items );
      lTsmScenario.Items.Add( oTsmScenItem );
      lTsmScenario.Snapshot = this.Snapshot;

      return lTsmScenario;
    }

    public TsmScenario ExtendChoice( TsmScenItem oTsmScenItem )
    {
      TsmScenario lTsmScenario = new TsmScenario( this.Parent );

      lTsmScenario.Items = new List<TsmScenItem>( this.Items );
      lTsmScenario.Items.Add( oTsmScenItem );
      lTsmScenario.Snapshot = this.Parent.Snapshot;

      return lTsmScenario;
    }

    public void Add( TsmScenItem oTsmScenItem )
    {
      Items.Add( oTsmScenItem );
    }

    public void Update( TsmSnapshot oSnapshot )
    {
      oSnapshot.Parent = this.Snapshot;
      this.Snapshot = oSnapshot;
    }

    public bool HasCycle()
    {
      TsmSnapshot lSnapshot = this.Snapshot;

      while ( null != (lSnapshot = lSnapshot.Parent) )
      {
        if ( lSnapshot.Equals( this.Snapshot ) )
        {
          return true;
        }
      }
      return false;
    }

    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      int i = 0;
      foreach ( TsmScenItem tItem in Items )
      {
        sb.Append( string.Format( "{0,3} ", i++ ) );
        sb.AppendLine( tItem.ToString() );
      }
      return sb.ToString();
    }
  }

  public class TsmSnapshot
  {
    public TsmSnapshot  Parent = null;
    public string       SnapClasses;
    public string       SnapCbThread;
    public string       SnapThreads;
    private int         MyHash = 0;

    public TsmSnapshot( string sSnapClasses, string sSnapCbThread, string sSnapThreads )
    {
      SnapClasses  = sSnapClasses;
      SnapCbThread = sSnapCbThread;
      SnapThreads  = sSnapThreads;

      MyHash = ToString().GetHashCode();
    }

    public override string ToString()
    {
      return "{" + SnapClasses + " &" + SnapCbThread + " &" + SnapThreads + "}";
    }

    public override bool Equals( Object o2 )
    {
      TsmSnapshot ss2 = (TsmSnapshot) o2;
      return (this.SnapClasses == ss2.SnapClasses) && (this.SnapCbThread == ss2.SnapCbThread) && (this.SnapThreads == ss2.SnapThreads);
    }

    public override int GetHashCode()
    {
      return MyHash;
    }
  }

  public class TsmSnapdata
  {
    public int          NrRecords;
    public int          NrCalls;
    public TsmScenario  Scenario;

    public TsmSnapdata( int iNrRecords, int iNrCalls, TsmScenario oScenario )
    {
      NrRecords = iNrRecords;
      NrCalls = iNrCalls;
      Scenario = oScenario;
    }
  }

  public class TsmRzVousSem
  {
    private Semaphore mSem;
    private bool      mbRendezVous = false;

    public TsmRzVousSem()
    {
      mSem = new Semaphore( 0, 1 );
    }

    public void WaitOne()
    {
      //Tsm.Assert( ! mbRendezVous );
      mbRendezVous = true;
      mSem.WaitOne();
      mbRendezVous = false;
    }

    public void Release()
    {
      //Tsm.Assert( mbRendezVous );
      mSem.Release();
    }

    public void Reset()
    {
      if ( mbRendezVous )
      {
        mbRendezVous = false;
        // Easiest way to reset the semaphore.
        mSem = new Semaphore( 0, 1 );
      }
    }
  }

  public class TsmTriggerSem
  {
    private Semaphore mSem;

    public TsmTriggerSem()
    {
      mSem = new Semaphore( 0, 1 );
    }

    public void WaitOne()
    {
      mSem.WaitOne();
    }

    public void Release()
    {
      try
      {
        mSem.Release();
      }
      catch ( SemaphoreFullException )
      { /*
         * Allowed to prevent excessive trigger looping.
         */
      }
    }
  }

  [Synchronization]
  public class TsmQueue : Queue<TsmEnvelope>
  {
    private int mMaxQueueSize = 0;
    private int mSeqId;

    public int MaxQueueSize
    {
      get { return mMaxQueueSize; }
    }
  
    public int SeqId
    {
      get { return mSeqId; }
      set { mSeqId = value; }
    }

    public void Put( TsmEnvelope oEnvelope )
    {
      if ( ! TsmSupervisor.QueueOverflow( this.Count ) )
      {
        base.Enqueue( oEnvelope );
        mMaxQueueSize = Math.Max( mMaxQueueSize, this.Count );
      }
    }

    public new TsmEnvelope Dequeue()
    {
      if ( 0 == this.Count )
      {
        return null;
      }

      return base.Dequeue();
    }

    public string Snap()
    {
      StringBuilder sb = new StringBuilder();

      foreach ( TsmEnvelope tEnv in this )
      {
        sb.Append( tEnv.Snap() + "," );
      }

      return sb.ToString();
    }
  }

  [Synchronization]
  public class TsmExtQueue : TsmQueue
  {
  }

  [Synchronization]
  public class TsmIntQueue : TsmQueue
  {
  }

  public enum TsmThreadState
  {
    RUNNING = 0,
    WAITING_MSG,
    WAITING_RV,
    WAITING_RV2,
    INTERRUPTED
  }

  [Synchronization]
  public class TsmThread
  {
    public TsmThreadType    mType;
    public int              mSequence;
    protected Thread        mThread;
    protected TsmExtQueue   mExternalQueue = new TsmExtQueue();
    protected TsmIntQueue   mInternalQueue;
    protected Mutex         mExclusiveMutex = new Mutex();
    protected bool          mbExclusive = false;
    protected TsmRzVousSem  mRendezVousSem = new TsmRzVousSem();
    protected TsmTriggerSem mTriggerSem = new TsmTriggerSem();
    protected Thread        mRendezVousThread;

    public TsmThread() : this( TsmThreadType.THREAD )
    {
    }

    public TsmThread( TsmThreadType Type )
    {
      mType = Type;
      mInternalQueue = new TsmIntQueue();
    }

    public Thread Thread
    {
      get { return mThread; }
    }

    public TsmExtQueue ExtQueue
    {
      get { return mExternalQueue; }
    }

    public TsmIntQueue IntQueue
    {
      get { return mInternalQueue; }
    }

    public int MaxExtQueueSize
    {
      get { return mExternalQueue.MaxQueueSize; }
    }

    public int MaxIntQueueSize
    {
      get { return mInternalQueue.MaxQueueSize; }
    }

    public bool HasMessage
    {
      get { return ( mExternalQueue.Count + mInternalQueue.Count > 0 ); }
    }

    public string MaxQueueSizes()
    {
      return string.Format( "{0,3}  {1,3}  {2}"
        , this.MaxExtQueueSize
        , this.MaxIntQueueSize
        , this.Thread.Name );
    }

    public virtual void Prepare( string sName )
    {
      mThread = new Thread( TsmThread.Worker );
      mSequence = TsmSupervisor.Register( this, mType );
      mExternalQueue.SeqId = mSequence;
      mInternalQueue.SeqId = mSequence;
      mThread.Name = "Tism" + mSequence + "_" + sName;

      TsmSupervisor.Announce();
    }

    public virtual void Run()
    {
      mThread.Start( this );
    }

    public virtual void Halt()
    {
      mThread.Abort();
      mThread.Join();
    }
    
    public virtual void Interrupt()
    {
      mThread.Interrupt();
    }

    public virtual void Cease()
    {
    }

    public void ExclusiveLock()
    {
      //Tsm.Assert( ! mbExclusive );
      mExclusiveMutex.WaitOne();
      mbExclusive = true;
    }

    public void ExclusiveUnlock()
    {
      mbExclusive = false;
      mExclusiveMutex.ReleaseMutex();
    }

    public void ExclusiveReset()
    {
      mbExclusive = false;
      mExclusiveMutex = new Mutex();
    }

    public void RendezVousWait()
    {
      mRendezVousThread = Thread.CurrentThread;
      mRendezVousSem.WaitOne();
    }

    public void RendezVousNotify()
    {
      Tsm.Seq( "R," + this.mExternalQueue.SeqId );
      TsmSupervisor.SetState( mRendezVousThread );
      mRendezVousSem.Release();
    }

    public void TriggerWait()
    {
      mTriggerSem.WaitOne();
    }

    public void TriggerNotify()
    {
      mTriggerSem.Release();
    }

    public void ASyncCall( TsmEnvelope oEnvelope )
    {
      mExternalQueue.Put( oEnvelope );
    }

    public void SyncCall( TsmEnvelope oEnvelope )
    {
      ExclusiveLock();

      oEnvelope.Trigger.RvState = TsmRvState.WAIT;
      mExternalQueue.Put( oEnvelope );

      TsmSupervisor.SetState( TsmThreadState.WAITING_RV );
      RendezVousWait();
      TsmSupervisor.SetState( TsmThreadState.WAITING_RV2 );

      ExclusiveUnlock();
    }

    protected TsmRvState Handle( TsmEnvelope oEnvelope )
    {
      oEnvelope.Target.ExecuteTrigger( oEnvelope.Trigger );
      if ( oEnvelope.Trigger.RvState == TsmRvState.READY )
      {
        RendezVousNotify();
        oEnvelope.Trigger.RvState = TsmRvState.IDLE;
      }
      return oEnvelope.Trigger.RvState;
    }

    public static void Worker( object oObject )
    {
      TsmThread   oTsmThread = (TsmThread) oObject;
      TsmEnvelope lEnvelope;
      TsmRvState  lRvState = TsmRvState.IDLE;

      for ( ; ; )
      {
        // Wait for message controlled by Supervisor
        TsmSupervisor.SetState( TsmThreadState.WAITING_MSG );

        lEnvelope = oTsmThread.mInternalQueue.Dequeue();
        if ( null != lEnvelope )
        {
          lEnvelope.Trigger.RvState = lRvState;

          if ( !lEnvelope.Trigger.IsForeign )
          {
            Tsm.Seq( "I," + oTsmThread.mSequence );
          }
          try
          {
            lRvState = oTsmThread.Handle( lEnvelope );
          }
          catch( ThreadInterruptedException )
          { /*
             * Escape from processing.
             */
            lRvState = TsmRvState.IDLE;
            TsmSupervisor.SetState( TsmThreadState.INTERRUPTED );
          }
          catch ( Exception e )
          {
            Tsm.ShowException( "@S", e );
          }
        }
        else
        {
          lEnvelope = oTsmThread.mExternalQueue.Dequeue();
          if ( null != lEnvelope )
          {
            if ( lEnvelope.Trigger.RvState == TsmRvState.WAIT )
            {
              lRvState = TsmRvState.WAIT;
            }
            else
            {
              lEnvelope.Trigger.RvState = lRvState;
            }

            if ( !lEnvelope.Trigger.IsForeign )
            {
              Tsm.Seq( "E," + oTsmThread.mSequence );
            }
            try
            {
              lRvState = oTsmThread.Handle( lEnvelope );
            }
            catch( ThreadInterruptedException )
            { /*
               * Escape from processing.
               */
              lRvState = TsmRvState.IDLE;
              TsmSupervisor.SetState( TsmThreadState.INTERRUPTED );
            }
            catch ( Exception e )
            {
              Tsm.ShowException( "@Q", e );
            }
          }
        }
      }
    }

    public void TsmReset1()
    {
      mExternalQueue.Clear();
      mInternalQueue.Clear();
    }

    public void TsmReset2()
    {
      mRendezVousSem.Reset();
      ExclusiveReset();
    }

    public string Snap()
    {
      StringBuilder sb = new StringBuilder();

      if ( mExternalQueue.Count > 0 )
      {
        sb.Append( " Q:" );
      }
      sb.Append( mExternalQueue.Snap() );

      if ( mInternalQueue.Count > 0 )
      {
        sb.Append( " S:" );
      }
      sb.Append( mInternalQueue.Snap() );

      return sb.ToString();
    }
  }

  [Synchronization]
  public class TsmCbThread : TsmThread
  {
    public TsmCbThread() : base( TsmThreadType.CB_THREAD )
    {
    }

    public void Prepare()
    {
      mThread = new Thread( TsmCbThread.Worker );
      mSequence = TsmSupervisor.Register( this, mType );
      mExternalQueue.SeqId = mSequence;
      mInternalQueue.SeqId = mSequence;
      mThread.Name = "TismRTE";
    }

    public void NativeCall( TsmEnvelope oEnvelope )
    {
      mExternalQueue.Put( oEnvelope );
    }

    protected void CbHandle( TsmEnvelope oEnvelope )
    {
      oEnvelope.Target.ExecuteNative( oEnvelope.Trigger.Command );
    }

    public new static void Worker( object oObject )
    {
      TsmCbThread oTsmThread = (TsmCbThread) oObject;
      TsmEnvelope lEnvelope;

      for ( ; ; )
      {
        // Wait for message controlled by Supervisor
        TsmSupervisor.SetState( TsmThreadState.WAITING_MSG );

        lEnvelope = oTsmThread.mExternalQueue.Dequeue();
        if ( null != lEnvelope )
        {
          if ( !lEnvelope.Trigger.IsForeign )
          {
            Tsm.Seq( "F" );
          }
          try
          {
            oTsmThread.CbHandle( lEnvelope );
          }
          catch( ThreadInterruptedException )
          { /*
             * Escape from processing.
             */
            TsmSupervisor.SetState( TsmThreadState.INTERRUPTED );
            Tsm.Lgl( 5, "@q : escape" );
          }
          catch ( Exception e )
          {
            Tsm.ShowException( "@q", e );
          }
        }
        else
        {
          lEnvelope = oTsmThread.mInternalQueue.Dequeue();
          if ( null != lEnvelope )
          {
            if ( !lEnvelope.Trigger.IsForeign )
            {
              Tsm.Seq( "J" );
            }
            try
            {
              oTsmThread.Handle( lEnvelope );
            }
            catch( ThreadInterruptedException )
            { /*
               * Escape from processing.
               */
              TsmSupervisor.SetState( TsmThreadState.INTERRUPTED );
              Tsm.Lgl( 5, "@s : escape" );
            }
            catch ( Exception e )
            {
              Tsm.ShowException( "@s", e );
            }
          }
        }
      }
    }
  }

  public class TsmActivity
  {
    private int         mNrAnnounce = 0;
    private int         mNrAbandon = 0;
    private Semaphore   mSem = new Semaphore( 0, 16 );

    public void Wait()
    {
      while ( mNrAnnounce != mNrAbandon )
      {
        mSem.WaitOne();
      }
    }

    public void Announce()
    {
      mNrAnnounce++;
    }

    public void Abandon()
    {
      mNrAbandon++;
      mSem.Release();
    }
  }

  public enum TsmThreadType
  {
    SUPERVISOR,
    CB_THREAD,
    THREAD
  }

  public class TsmThreadData
  {
    public TsmThreadType  Type;
    public int            ThreadId;
    public TsmThread      TsmThread;
    public TsmThreadState State;

    public bool HasMessage
    {
      get { return TsmThread.HasMessage; }
    }
  }

  /*
   * Received client natives have negative values.
   */
  public class TsmNativeLists
  {
    public List<int>  ClientNatives = new List<int>();
    public List<int>  ServerNatives = new List<int>();
  }

  [Synchronization]
  public class TsmSupervisor
  {
    private static TsmTriggerSem mTriggerSem = new TsmTriggerSem();
    private static TsmActivity   mActivity;
    private static int           mThreadId = 0;
    private static TsmCbThread   mCbThread;
    private static TsmThreadData MyThreadData;
    private static TsmThreadData CbThreadData;
    private static Dictionary<Thread,TsmThreadData> mThreadDatas = new Dictionary<Thread,TsmThreadData>();

    public static void Initialise()
    {
      Tsm.Open();
      mActivity = new TsmActivity();
      Register( null, TsmThreadType.SUPERVISOR );

      mCbThread = new TsmCbThread();
      mCbThread.Prepare();
      mCbThread.Run();
    }

    public static void Terminate()
    {
      mCbThread.Halt();
      mCbThread.Cease();
      Tsm.Close();
    }

    public static TsmIntQueue GetIntQueue()
    {
      return mCbThread.IntQueue;
    }

    public static void Announce()
    {
      mActivity.Announce();
    }

    public static void Abandon()
    {
      mActivity.Abandon();
    }

    public static int Register( TsmThread oTsmThread )
    {
      return Register( oTsmThread, TsmThreadType.THREAD );
    }

    public static int Register( TsmThread oTsmThread, TsmThreadType eType )
    {
      TsmThreadData lThreadData;
      Thread oThread;
      
      if ( null != oTsmThread )
      {
        oThread = oTsmThread.Thread;
      }
      else
      {
        oThread = Thread.CurrentThread;
      }

      if ( mThreadDatas.ContainsKey( oThread ) )
      {
        throw new TsmException( "TismBUG: Doubly registered thread" );
      }

      lThreadData = new TsmThreadData();
      lThreadData.Type      = eType;
      lThreadData.ThreadId  = mThreadId++;
      lThreadData.TsmThread = oTsmThread;
      lThreadData.State     = TsmThreadState.RUNNING;

      switch ( lThreadData.Type )
      {
      case TsmThreadType.SUPERVISOR:
        MyThreadData = lThreadData;
        break;
      case TsmThreadType.CB_THREAD:
        CbThreadData = lThreadData;
        break;
      case TsmThreadType.THREAD:
        mThreadDatas[ oThread ] = lThreadData;
        break;
      }
     
      return lThreadData.ThreadId;
    }

    public static void SetState( Thread oThread )
    {
      SetState( oThread, TsmThreadState.RUNNING );
    }

    public static void SetState( TsmThreadState eState )
    {
      SetState( Thread.CurrentThread, eState );
    }

    public static void SetState( Thread lThread, TsmThreadState eState )
    {
      TsmThreadData lThreadData;

      if ( lThread == CbThreadData.TsmThread.Thread )
      {
        lThreadData = CbThreadData;
      }
      else
      {
        if ( ! mThreadDatas.ContainsKey( lThread ) )
        {
          throw new TsmException( "TismBUG: Unregistered thread" );
        }
        lThreadData = mThreadDatas[ lThread ];
      }

      lThreadData.State = eState;
      switch ( lThreadData.Type )
      {
      case TsmThreadType.SUPERVISOR:
        switch ( lThreadData.State )
        {
        case TsmThreadState.RUNNING:
        case TsmThreadState.INTERRUPTED:
          throw new TsmException( "TismBUG: Supervisor should not report" );
          //break;
        case TsmThreadState.WAITING_MSG:
        case TsmThreadState.WAITING_RV:
        case TsmThreadState.WAITING_RV2:
          throw new TsmException( "TismBUG: Supervisor should not wait" );
          //break;
        }
        break;
      case TsmThreadType.CB_THREAD:
      case TsmThreadType.THREAD:
        switch ( lThreadData.State )
        {
        case TsmThreadState.RUNNING:
          break;
        case TsmThreadState.WAITING_MSG:
          mTriggerSem.Release();
          Pause( lThreadData );
          break;
        case TsmThreadState.WAITING_RV:
          mTriggerSem.Release();
          break;
        case TsmThreadState.WAITING_RV2:
          mTriggerSem.Release();
          Pause( lThreadData );
          break;
        case TsmThreadState.INTERRUPTED:
          lThreadData.State = TsmThreadState.RUNNING;
          break;
        }
        break;
      }
    }

    private static void CheckAllThreadsWaiting()
    {
      while ( ! AllThreadsWaiting() )
      {
        mTriggerSem.WaitOne();
      }

      while ( ! AllThreadsSleeping() )
      {
        Thread.Sleep( 1 );
      }
    }

    private static bool AllThreadsWaiting()
    {
      if ( TsmThreadState.RUNNING == CbThreadData.State )
      {
        return false;
      }

      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        if ( TsmThreadState.RUNNING == lThreadData.State )
        {
          return false;
        }
      }
      return true;
    }

    private static bool AllThreadsSleeping()
    {
      if ( CbThreadData.TsmThread.Thread.ThreadState != System.Threading.ThreadState.WaitSleepJoin )
      {
        return false;
      }

      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        if ( lThreadData.TsmThread.Thread.ThreadState != System.Threading.ThreadState.WaitSleepJoin )
        {
          return false;
        }
      }
      return true;
    }

    private static string GetThreadStates( string sHeader )
    {
      StringBuilder sb = new StringBuilder();
      TsmThreadData lThreadData;

      sb.AppendLine( sHeader );
      sb.AppendLine( "Thread " + MyThreadData.ThreadId
        + " Type " + MyThreadData.Type
        + " State " + MyThreadData.State );
      sb.AppendLine( "Thread " + CbThreadData.ThreadId
        + " Type " + CbThreadData.Type
        + " State " + CbThreadData.State );
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        lThreadData = kvp.Value;
        sb.AppendLine( "Thread " + lThreadData.ThreadId
          + " Type " + lThreadData.Type
          + " State " + lThreadData.State );
      }
      return sb.ToString();
    }
    
    private static Dictionary<TsmClass,TsmNativeLists> Natives = new Dictionary<TsmClass,TsmNativeLists>();

    private static void NativesReset()
    {
      Natives.Clear();
    }

    public static void NativeEntry( TsmClass oOwner, int [] aNatives )
    {
      TsmNativeLists lLists;

      if ( ! Natives.ContainsKey( oOwner ) )
      {
        lLists = new TsmNativeLists();
        Natives[ oOwner ] = lLists;
      }
      else
      {
        lLists = Natives[ oOwner ];
      }

      foreach ( int n in aNatives )
      {
        if ( n < 0 )
        {
          lLists.ClientNatives.Add( n );
        }
        else
        {
          lLists.ServerNatives.Add( n );
        }
      }
    }

    public static void NativeExit( TsmClass oOwner, int [] aNatives )
    {
      TsmNativeLists lLists;

      if ( ! Natives.ContainsKey( oOwner ) )
      {
        throw new TsmException( "TismBUG: NativeExit " + oOwner );
      }

      lLists = Natives[ oOwner ];

      foreach ( int n in aNatives )
      {
        if ( n < 0 )
        {
          lLists.ClientNatives.Remove( n );
        }
        else
        {
          lLists.ServerNatives.Remove( n );
        }
      }
    }

    public static Dictionary<TsmSnapshot,TsmSnapdata>  Snapshots = new Dictionary<TsmSnapshot,TsmSnapdata>();
    public static List<TsmClass>        Classes = new List<TsmClass>();
    public static List<TsmScenario>     PastScenarios = new List<TsmScenario>();
    public static Stack<TsmScenario>    FutureScenarios = new Stack<TsmScenario>();
    public static TsmScenario           CurrentScenario;
    public static int                   CurrentScenItemIndex = 0;

    private static int          NumberOfDeadlocks = 0;
    private static int          NumberOfLivelocks = 0;
    private static int          NumberOfCycles = 0;
    private static long         NumberOfLoops = 0;
    private static DateTime     StartTime;
    private static DateTime     StopTime;

    public static string GetCurrentClasses()
    {
      StringBuilder sb = new StringBuilder();
      int c = 0;
      foreach( TsmClass tClass in Classes )
      {
        sb.Append( string.Format( "{0,3} ", ++c ) );
        sb.AppendLine( tClass.TsmSnap() );
      }
      return sb.ToString();
    }

    public static string GetCurrentScenario()
    {
      return PastScenarios.Last().ToString();
    }

    private static void Reset()
    {
      mCbThread.TsmReset1();
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        lThreadData.TsmThread.TsmReset1();
      }

      if ( TsmThreadState.WAITING_RV == CbThreadData.State )
      {
        Tsm.Lgl( 7, "Interrupt Cb" );
        CbThreadData.State = TsmThreadState.RUNNING;
        mCbThread.Interrupt();
      }
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        if ( TsmThreadState.WAITING_RV == lThreadData.State )
        {
          Tsm.Lgl( 7, "Interrupt T_" + lThreadData.TsmThread.mSequence );
          lThreadData.State = TsmThreadState.RUNNING;
          lThreadData.TsmThread.Interrupt();
        }
      }
      CheckAllThreadsWaiting();

      mCbThread.TsmReset2();
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        lThreadData.TsmThread.TsmReset2();
      }

      foreach ( TsmClass tClass in Classes )
      {
        tClass.TsmReset1();
      }
      NativesReset();
    }

    private static void Pause( TsmThreadData oThreadData )
    {
      oThreadData.TsmThread.TriggerWait();
    }

    private static void Continue( TsmThreadData oThreadData )
    {
      oThreadData.State = TsmThreadState.RUNNING;
      oThreadData.TsmThread.TriggerNotify();
    }

    private static void Replay()
    {
      foreach ( TsmClass tClass in Classes )
      {
        tClass.TsmReset2();
      }

      CurrentScenItemIndex = 0;
      TsmRecorder.Reset();

      for ( ; ; )
      {
        CheckAllThreadsWaiting();

        if ( CurrentScenItemIndex == CurrentScenario.Items.Count )
        { /*
           * Ready replaying, possibly due to choice items, or
           * empty scenario.
           */
          TsmRecorder.Enable();
          break;
          /****/
        }
        if ( CurrentScenItemIndex == CurrentScenario.Items.Count-1 )
        { /*
           * The last item is 'new'.
           */
          TsmRecorder.Enable();
        }
        ReplayScenItem( CurrentScenario.Items[ CurrentScenItemIndex++ ] );
      }
    }

    private static void ReplayNative( TsmScenItem oItem )
    {
      Tsm.Lgl( 6, "Natv: " + oItem.Index + " @ " + oItem.Receiver.FullName() );
      if ( oItem.Index < 0 )
      { /*
         * Client native.
         * Executed by CbThread.
         */
        TsmTrigger lTrigger = new TsmTrigger( -oItem.Index );
        TsmEnvelope lEnvelope = new TsmEnvelope( lTrigger, oItem.Receiver );
        mCbThread.NativeCall( lEnvelope );
        Continue( CbThreadData );
      }
      else
      { /*
         * Server native.
         * Executed by Supervisor.
         */
        oItem.Receiver.ExecuteNative( oItem.Index );
      }
    }

    /*
     * The state changes :
     *     Item.ThreadData.State = TsmThreadState.RUNNING;
     * are needed to force the Supervisor to wait for the threads.
     */
    private static void ReplayScenItem( TsmScenItem Item )
    {
      switch ( Item.Kind )
      {
      case TsmScenItemKind.RELEASE_CB_MSG :
        //Tsm.Assert( Item.ThreadData.HasMessage );
        Continue( Item.ThreadData );
        break;
      case TsmScenItemKind.RELEASE_CB_RV2 :
        //Tsm.Assert( Item.ThreadData.State == TsmThreadState.WAITING_RV2 );
        Continue( Item.ThreadData );
        break;
      case TsmScenItemKind.RELEASE_MSG :
        //Tsm.Assert( Item.ThreadData.HasMessage );
        Continue( Item.ThreadData );
        break;
      case TsmScenItemKind.RELEASE_RV2 :
        //Tsm.Assert( ( Item.ThreadData.State & TsmThreadState.WAITING_RV ) != 0 );
        Continue( Item.ThreadData );
        break;
      case TsmScenItemKind.NATIVE :
        ReplayNative( Item );
        break;
      case TsmScenItemKind.CHOICE :
        TsmRecorder.Enable();
        TsmRecorder.Show( "**** BUG " );
        throw new TsmException( "TismBUG: Handled by 'ChoiceMax'" );
        //break;
      }
    }

    private static TsmScenItem GetThreadDataRV2()
    {
      TsmScenItem lScenItem;

      if ( TsmThreadState.WAITING_RV2 == CbThreadData.State )
      {
        lScenItem = new TsmScenItem();
        lScenItem.Kind = TsmScenItemKind.RELEASE_CB_RV2;
        lScenItem.ThreadData = CbThreadData;
        return lScenItem;
      }

      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        if ( TsmThreadState.WAITING_RV2 == lThreadData.State )
        {
          lScenItem = new TsmScenItem();
          lScenItem.Kind = TsmScenItemKind.RELEASE_RV2;
          lScenItem.ThreadData = lThreadData;
          return lScenItem;
        }
      }
      
      return null;
    }

    private static void ReplayExtend( TsmScenario oScenario )
    {
      TsmScenItem lScenItem;

      while ( null != (lScenItem = GetThreadDataRV2()) )
      {
        oScenario.Add( lScenItem );
        ReplayScenItem( lScenItem );
        CurrentScenItemIndex++;
        CheckAllThreadsWaiting();
      }
    }

    private static bool     mbQueueOverflow = false;
    private static int      mMaximumQueueSize = 3;
    /*
     * Called by Queues.
     */
    public static bool QueueOverflow( int QueueCount )
    {
      if ( QueueCount >= mMaximumQueueSize )
      {
        TsmScenItem lScenItem = new TsmScenItem();
        lScenItem.Kind = TsmScenItemKind.QO;
        CurrentScenario.Add( lScenItem );
        CurrentScenItemIndex++;
        mbQueueOverflow = true;
      }
      return mbQueueOverflow;
    }

    private static bool     mbVerify = true;
    private static Random   mRandom = new Random( 1 );
    private static int      mRandomSeed = 1;
    private static long     mMaxNrSimLoops = 50;
    /*
     * Called by Fsms.
     */
    public static bool ChoiceOK( int MaxNrChoices, int ChoiceIndex, string sGuard )
    {
      if ( mbVerify )
      {
        return ChoiceOKVerify( MaxNrChoices, ChoiceIndex, sGuard );
      }

      return ChoiceOKSimulate( MaxNrChoices, ChoiceIndex );
    }

    private static bool ChoiceOKVerify( int MaxNrChoices, int ChoiceIndex, string sGuard )
    {
      TsmScenItem lScenItem;

      if ( CurrentScenItemIndex < CurrentScenario.Items.Count )
      { /*
         * Still replaying.
         */
        lScenItem = CurrentScenario.Items[ CurrentScenItemIndex ];
        //Tsm.Assert( lScenItem.Kind == TsmScenItemKind.CHOICE );
        if ( lScenItem.Index != ChoiceIndex )
        {
          return false;
          /***********/
        }

        CurrentScenItemIndex++;
        if ( CurrentScenItemIndex == CurrentScenario.Items.Count )
        { /*
           * 'Execution' of the last item is happening now.
           */
          TsmRecorder.Enable();
        }
      }
      else
      { /*
         * The scenario has been replayed fully, and
         * now encounters a Choice node.
         */
        //Tsm.Assert( ChoiceIndex == 1 );
        TsmScenario lNewScenario;
        for ( int i=2; i<=MaxNrChoices; i++ )
        {
          lScenItem = new TsmScenItem();
          lScenItem.Kind = TsmScenItemKind.CHOICE;
          lScenItem.Index = i;
          lNewScenario = CurrentScenario.ExtendChoice( lScenItem );
          FutureScenarios.Push( lNewScenario );
        }

        lScenItem = new TsmScenItem();
        lScenItem.Kind = TsmScenItemKind.CHOICE;
        lScenItem.Index = 1;
        CurrentScenario.Add( lScenItem );
        CurrentScenItemIndex++;
      }
      Tsm.Seq( "H," + sGuard );
      return true;
    }

    public static void Verify()
    {
      List<TsmScenItem> lScenItems = new List<TsmScenItem>();
      bool              bProxyProtocol = false;
      TsmSnapshot       lSnapshot;
      TsmSnapdata       lSnapdata;

      mbVerify = true;
      StartTime = DateTime.Now;

      Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
      Process.GetCurrentProcess().ProcessorAffinity = (System.IntPtr)1;

      if ( Classes.Count == 2 )
      {
        if ( ( ( Classes[0] is TsmProxy ) && ( Classes[1] is TsmProxyUser ) )
          || ( ( Classes[1] is TsmProxy ) && ( Classes[0] is TsmProxyUser ) ) )
        {
          bProxyProtocol = true;
        }
      }
      CurrentScenario = new TsmScenario();
      FutureScenarios.Push( CurrentScenario );

      bool mbCurrent = false;
      for ( ; ; )
      {
        if ( FutureScenarios.Count == 0 )
        {
          // Console.WriteLine( "No more scenarios" );
          break;
        }
        CurrentScenario = FutureScenarios.Pop();
        PastScenarios.Add( CurrentScenario );

        if ( mbCurrent )
        {
          mbCurrent = false;

          Tsm.Seq( "S," + CurrentScenario.Id + "," + CurrentScenario.ParentId );
          Tsm.Lgl( 2, "Scenario [ " + CurrentScenario.Id + "  :  " + CurrentScenario.ParentId + " ] -------------------------------------------------------" );
          Tsm.Lgl( 3, CurrentScenario.ToString() );
          if ( CurrentScenario.Items.Count > 0 )
          {
            CheckAllThreadsWaiting();
            ReplayScenItem( CurrentScenario.Items[ CurrentScenItemIndex++ ] );
            //Tsm.Assert( CurrentScenItemIndex == CurrentScenario.Items.Count );
          }
          CheckAllThreadsWaiting();
        }
        else
        {
          Tsm.Seq( "N" );
          Reset();
          Tsm.Seq( "S," + CurrentScenario.Id + "," + CurrentScenario.ParentId );
          Tsm.Lgl( 2, "Scenario [ " + CurrentScenario.Id + "  :  " + CurrentScenario.ParentId + " ] -------------------------------------------------------" );
          Tsm.Lgl( 3, CurrentScenario.ToString() );
          Replay();
        }
        ReplayExtend( CurrentScenario );
        /*
         * Now, no threads have state RV2 !
         */

        lSnapshot = new TsmSnapshot(
          SnappedClasses(),
          CbThreadData.TsmThread.Snap(),
          SnappedThreads() );
        CurrentScenario.Update( lSnapshot );

        if ( mbQueueOverflow )
        {
          mbQueueOverflow = false;
          TsmRecorder.QueueOverflow();
          continue;
          /*******/
        }

        Tsm.Seq( "T," + lSnapshot.SnapClasses );
        if ( Snapshots.TryGetValue( lSnapshot, out lSnapdata ) )
        {
          Tsm.Lgl( 5, "Equal: " + lSnapdata.Scenario.Id );
          Tsm.Seq( "Q," + lSnapdata.Scenario.Id );
          if ( CurrentScenario.HasCycle() )
          {
            NumberOfCycles++;
            Tsm.Lgl( 4, "Cycle: -------------------------------------------------------" );
            Tsm.Seq( "Y" );
            if ( lSnapdata.NrCalls == TsmRecorder.NumberOfExtCalls )
            { /*
               * The number of external calls retrieved from the previous
               * position in lSnapdata is equal to the current number of
               * external calls in the Recorder.
               */
              NumberOfLivelocks++;
              TsmRecorder.Show( "**** LIVELOCK ", lSnapdata.NrRecords );
            }
          }
        }
        else
        {
          Tsm.Lgl( 5, "Rply: -------------------------------------------------------" );
          Tsm.Lgl( 5, "Snap:" + lSnapshot.ToString() );
          lSnapdata = new TsmSnapdata( TsmRecorder.NumberOfRecords, TsmRecorder.NumberOfExtCalls, CurrentScenario );
          Snapshots.Add( lSnapshot, lSnapdata );

          bool bClient = (CbThreadData.State == TsmThreadState.WAITING_MSG);
          lScenItems.Clear();
          AddScenarioCbThread( lScenItems );
          if ( bProxyProtocol )
          { /*
             * In case of a proxy protocol, native triggers are only
             * allowed in the absense of pending callbacks.
             */
            if ( lScenItems.Count == 0 )
            {
              AddScenarioNatives( lScenItems, bClient );
              AddScenarioThreads( lScenItems );
            }
          }
          else
          {
            AddScenarioNatives( lScenItems, bClient );
            AddScenarioThreads( lScenItems );
          }

          if ( lScenItems.Count == 0 )
          {
            NumberOfDeadlocks++;
            TsmRecorder.Show( "**** DEADLOCK " );
          }
          else
          {
            AddToCurrentScenario( lScenItems );
            mbCurrent = true;
          }
        }
      }

      /*
       * Cleanup pending threads.
       */
      Reset();
      StopTime = DateTime.Now;
    }

    public static void ConfigureSimulation( int iSeed, long iNrLoops )
    {
      mRandomSeed = iSeed;
      mMaxNrSimLoops = iNrLoops;
    }

    private static bool ChoiceOKSimulate( int MaxNrChoices, int ChoiceIndex )
    {
      return (mRandom.Next( MaxNrChoices + 1 - ChoiceIndex ) == 0);
    }

    public static void Simulate()
    {
      List<TsmScenItem> lScenItems = new List<TsmScenItem>();
      bool              bProxyProtocol = false;

      mbVerify = false;
      mRandom = new Random( mRandomSeed );
      StartTime = DateTime.Now;

      Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
      Process.GetCurrentProcess().ProcessorAffinity = (System.IntPtr)1;

      if ( Classes.Count == 2 )
      {
        if ( ( ( Classes[0] is TsmProxy ) && ( Classes[1] is TsmProxyUser ) )
          || ( ( Classes[1] is TsmProxy ) && ( Classes[0] is TsmProxyUser ) ) )
        {
          bProxyProtocol = true;
        }
      }

      bool bCurrent = false;
      NumberOfLoops = 0;
      do
      {
        if ( ++NumberOfLoops == 0 )
        {
          NumberOfLoops = 1;
        }

        Tsm.Lgl( 3, "Loop: " + NumberOfLoops );
        if ( bCurrent )
        {
          bCurrent = false;

          CheckAllThreadsWaiting();
          int r = mRandom.Next( lScenItems.Count );
          CurrentScenario.Extend( lScenItems[ r ] );
          ReplayScenItem( lScenItems[ r ] );
          CheckAllThreadsWaiting();
        }
        else
        {
          Reset();
          CurrentScenario = new TsmScenario();
          Replay();
        }

        TsmScenItem lScenItem;
        while ( null != ( lScenItem = GetThreadDataRV2() ) )
        {
          ReplayScenItem( lScenItem );
          CheckAllThreadsWaiting();
        }

        bool bClient = ( CbThreadData.State == TsmThreadState.WAITING_MSG );
        lScenItems.Clear();
        AddScenarioCbThread( lScenItems );
        if ( bProxyProtocol )
        { /*
           * In case of a proxy protocol, native triggers are only
           * allowed in the absense of pending callbacks.
           */
          if ( lScenItems.Count == 0 )
          {
            AddScenarioNatives( lScenItems, bClient );
            AddScenarioThreads( lScenItems );
          }
        }
        else
        {
          AddScenarioNatives( lScenItems, bClient );
          AddScenarioThreads( lScenItems );
        }

        if ( lScenItems.Count == 0 )
        {
          NumberOfDeadlocks++;
          TsmRecorder.Show( "**** DEADLOCK " );
        }
        else
        {
          //AddToCurrentScenario( lScenItems );
          bCurrent = true;
        }
      } while ( NumberOfLoops != mMaxNrSimLoops );

      Reset();
      StopTime = DateTime.Now;
    }

    public static string SnappedClasses()
    {
      StringBuilder sb = new StringBuilder();
      int c = 0;
      foreach( TsmClass tClass in Classes )
      {
        sb.Append( "" + ++c  );
        sb.Append( tClass.TsmSnap() );
      }
      return sb.ToString();
    }

    public static string SnappedThreads()
    {
      StringBuilder sb = new StringBuilder();
      int t = 0;
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        sb.Append( " T_" + ++t + ": " );
        sb.Append( lThreadData.TsmThread.Snap() );
      }
      return sb.ToString();
    }

    private static void AddToCurrentScenario( List<TsmScenItem> oScenItems )
    {
      TsmScenario lNewScenario;

      foreach( TsmScenItem si in oScenItems )
      {
        lNewScenario = CurrentScenario.Extend( si );
        FutureScenarios.Push( lNewScenario );
      }
    }

    private static void AddScenarioNatives( List<TsmScenItem> oScenItems, bool bClient )
    {
      foreach ( KeyValuePair<TsmClass,TsmNativeLists> kvp in Natives )
      {
        TsmNativeLists lLists = kvp.Value;
        TsmClass Receiver = kvp.Key;

        foreach ( int index in lLists.ServerNatives )
        {
          TsmScenItem lScenItem = new TsmScenItem();
          lScenItem.Kind = TsmScenItemKind.NATIVE;
          lScenItem.Index = index;
          lScenItem.Receiver = Receiver;
          oScenItems.Add( lScenItem );
        }

        if ( bClient )
        {
          foreach ( int index in lLists.ClientNatives )
          {
            TsmScenItem lScenItem = new TsmScenItem();
            lScenItem.Kind = TsmScenItemKind.NATIVE;
            lScenItem.Index = index;
            lScenItem.Receiver = Receiver;
            oScenItems.Add( lScenItem );
          }
        }
      }
    }

    private static void AddScenarioCbThread( List<TsmScenItem> oScenItems )
    {
      if ( CbThreadData.TsmThread.IntQueue.Count == 0 )
      {
        return;
      }

      if ( CbThreadData.State == TsmThreadState.WAITING_RV )
      {
        return;
      }
      //Tsm.Assert( CbThreadData.State == TsmThreadState.WAITING_MSG );

      TsmScenItem lScenItem = new TsmScenItem();
      lScenItem.Kind = TsmScenItemKind.RELEASE_CB_MSG;
      lScenItem.ThreadData = CbThreadData;
      oScenItems.Add( lScenItem );
    }

    private static void AddScenarioThreads( List<TsmScenItem> oScenItems )
    {
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        if ( TsmThreadState.WAITING_MSG == lThreadData.State )
        {
          if ( lThreadData.TsmThread.HasMessage )
          {
            TsmScenItem lScenItem = new TsmScenItem();
            lScenItem.Kind = TsmScenItemKind.RELEASE_MSG;
            lScenItem.ThreadData = lThreadData;
            oScenItems.Add( lScenItem );
          }
        }
      }
    }

    public static void Summary()
    {
      TimeSpan duration = StopTime - StartTime;
      Tsm.Log( "" );
      Tsm.Log( "SUMMARY" );
      Tsm.Log( "-------" );
      if ( mbVerify )
      {
        Tsm.Log( "Number of scenarios: " + PastScenarios.Count );
        Tsm.Log( "Number of cycles   : " + NumberOfCycles );
      }
      else
      {
        Tsm.Log( "Number of loops    : " + NumberOfLoops );
      }
      Tsm.Log( "Number of deadlocks: " + NumberOfDeadlocks );
      if ( mbVerify )
      {
        Tsm.Log( "Number of livelocks: " + NumberOfLivelocks );
      }
      Tsm.Log( "Number of unhandled functions  : " + TsmRecorder.NumberOfUnhandledFunctions );
      Tsm.Log( "Number of unvisited states     : " + TsmRecorder.NumberOfUnvisitedStates );
      Tsm.Log( "Number of unvisited transitions: " + TsmRecorder.NumberOfUnvisitedTransitions );
      Tsm.Log( "Number of rendez-vous errors   : " + TsmRecorder.NumberOfSyncReturnErrors );
      Tsm.Log( "Number of queue overflows      : " + TsmRecorder.NumberOfQueueOverflows );
      Tsm.Log( "Duration: " + duration.ToString().Substring( 0, 8 ) );
      Tsm.Log( "-------" );
      Tsm.Log( "Ext  Int  Thread's maximum queue sizes" );
      Tsm.Log( mCbThread.MaxQueueSizes() );
      foreach ( KeyValuePair<Thread,TsmThreadData> kvp in mThreadDatas )
      {
        TsmThreadData lThreadData = kvp.Value;
        Tsm.Log( lThreadData.TsmThread.MaxQueueSizes() );
      }
      Tsm.Log( "-------" );
    }
  }

  [Synchronization]
  public class TsmRecorder
  {
    public static int           NumberOfUnhandledFunctions = 0;
    public static int           NumberOfUnvisitedStates = 0;
    public static int           NumberOfUnvisitedTransitions = 0;
    public static int           NumberOfSyncReturnErrors = 0;
    public static int           NumberOfQueueOverflows = 0;
    
    private static List<string> Records;
    private static int          NrCalls;
    private static bool         bEnable;
    private static string       Separator = "============================================================";

    static TsmRecorder()
    {
      Reset();
    }

    public static int NumberOfRecords
    {
      get { return Records.Count; }
    }

    public static int NumberOfExtCalls
    {
      get { return NrCalls; }
    }

    public static void Reset()
    {
      Records = new List<string>();
      NrCalls = 0;
      bEnable = false;
    }

    public static void Enable()
    {
      bEnable = true;
    }

    public static void AddNative( string sNative )
    {
      string s = "NaTr: " + sNative;
      Records.Add( s );
    }

    public static void AddCall( string sCall, TsmClass oSource, TsmPort oPort )
    {
      TsmClass lTarget = oPort.Target;
      string sSource = "'foreign'";
      bool   bSourceIsMock = false;
      int    iSourceSeqId = -1;
      if ( ( oSource != null ) && ( oSource.SeqId > 0 ) )
      {
        sSource = oSource.FullName();
        bSourceIsMock = oSource.IsMock;
        iSourceSeqId = oSource.SeqId;
      }

      string s = "Call: " + sCall + " @ " + sSource + " -> " + lTarget.FullName();
      Records.Add( s );
      if ( bSourceIsMock || lTarget.IsMock )
      {
        NrCalls++;
      }
      if ( iSourceSeqId >= 0 )
      {
        Tsm.Seq( "C," + sCall + "," + iSourceSeqId + "," + lTarget.SeqId + "," + oPort.ToSeq() );
      }
    }

    public static void AddStates( string sStates )
    {
      string s = "States: " + sStates;
      Records.Add( s );
    }

    public static void UnhandledFunction( string sEvent )
    {
      NumberOfUnhandledFunctions++;
      Show( "**** UNHANDLED FUNCTION : " + sEvent );
    }

    public static void UnvisitedState( string sEvent )
    {
      StringBuilder sb = new StringBuilder();

      NumberOfUnvisitedStates++;
      sb.AppendLine( Separator );
      sb.AppendLine( "Unvisited state: " + sEvent );
      Tsm.Log( sb.ToString() );
    }

    public static void UnvisitedTransition( string sState, string sEvent )
    {
      StringBuilder sb = new StringBuilder();

      NumberOfUnvisitedTransitions++;
      sb.AppendLine( Separator );
      sb.AppendLine( "@ state: " + sState );
      sb.AppendLine( "Unvisited transition: " + sEvent );
      Tsm.Log( sb.ToString() );
    }

    public static void SyncReturnError( string sEvent )
    {
      NumberOfSyncReturnErrors++;
      Show( sEvent );
    }

    public static void QueueOverflow()
    {
      NumberOfQueueOverflows++;
      Show( "**** QUEUE OVERFLOW " );
    }

    public static void Show( string sEvent )
    {
      if ( bEnable )
      {
        StringBuilder sb = new StringBuilder();
        sb.AppendLine( Separator );
        sb.AppendLine( sEvent );

        foreach ( string s in Records )
        {
          sb.AppendLine( s );
        }

        sb.Append( TsmSupervisor.SnappedClasses() );
        Tsm.Log( sb.ToString() );
      }
      Tsm.Seq( "X," + sEvent );
    }

    public static void Show( string sEvent, int iNrRecords )
    {
      if ( bEnable )
      {
        StringBuilder sb = new StringBuilder();
        sb.AppendLine( Separator );
        sb.AppendLine( sEvent );

        for ( int i=0; i<iNrRecords; i++ )
        {
          sb.AppendLine( Records[i] );
        }
        sb.AppendLine( "=== start of cycle ===" );
        for ( int i=iNrRecords; i<Records.Count; i++ )
        {
          sb.AppendLine( Records[i] );
        }

        sb.AppendLine( "Classes:" );
        sb.Append( TsmSupervisor.GetCurrentClasses() );

        sb.AppendLine( "Scenario:" );
        sb.Append( TsmSupervisor.GetCurrentScenario() );

        Tsm.Log( sb.ToString() );
      }
      Tsm.Seq( "X," + sEvent );
    }
  }
}
